open Format
open Formatx
open Term
open Subst
open Overlap
open Rewriting
open Iroiro
open Arg
open Result

(* proof tree *)

type proof = 
  | KH12 of Rules.t (* R *) * Rules.t (* S *) * Rules.t (* CP_S(R) *)  * proof
  | HM11 of Rules.t (* R *) * Rules.t (* CPS'(R) *) * Rules.t (* CP(R) *)
  | VO08 of Rules.t (* R *) * (Rule.t * int) list (* rule label *)
  | JK86 of Rules.t (* R *) * Rules.t (* CP_AC(R) *)

let print_label ppf (rule, n) =
  fprintf ppf "[%a] = %d" Rule.print rule n 

let rec print_proof ppf = function
  | HM11 (rs, cps, cp) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Hirokawa and Middeldorp, 2011]@.\
         R:@.%a@.\
         CPS'(R):@.%a@.\
         CP(R):@.%a@."
        Rules.print rs
        Rules.print cps
        Rules.print cp
  | KH12 (rs, ss, ecp, proof) ->
      fprintf ppf 
        "---- CR(R cup S)@.\
         ---- [Klein and Hirokawa, 2012]@.\
         R:@.%a@.\
         S:@.%a@.\
         CP_S(R):@.%a@.\
         CR(S):@.see below.@.@.%a"
	Rules.print rs Rules.print ss
	Rules.print ecp print_proof proof
  | JK86 (rs, cp) ->
      fprintf ppf 
        "---- CR(R)@.\
         ---- [Jouannaud and Kirchner, 1986]@.\
         R:@.%a@.\
         CP_AC(R):@.%a"
	Rules.print rs Rules.print cp
  | VO08 (rs, a) ->
      fprintf ppf
        "---- CR(R)@.\
         ---- [van Oostrom, RTA 2008].@.\
         R:@.%a@.@.\
         Label:@.%a@.@."
        Rules.print rs (print_list print_label "@.") a

let print_disproof ppf (rs, s, t) =
  fprintf ppf
    "---- Non-confluence check@.\
     R:@.%a@.@.\
     Unjoinable conversion:@.%a <->* %a@.@." 
    Rules.print rs
    Term.print s Term.print t

(* Auciliary functions *)

exception Failed
exception Disproved of Term.t * Term.t

let rec try_find f = function
  | [] -> raise Failed
  | x :: xs -> try f x with Failed -> try_find f xs

let rec pairwise_find p = function
  | [] -> raise Failed
  | x :: xs ->
      try 
        let y = List.find (p x) xs in (x, y) 
      with Not_found -> pairwise_find p xs

(* Global variables *)

let filenames  = ref []
let t          = ref None 
let tt         = ref "ttt2"
let actt       = ref "muterm -i"
let smt        = ref "minismt -m"
let wcrk       = ref 5
let rlvk       = ref 5
let rlck       = ref 5

let rec tcap rs = function
  | V _ -> V (fresh ())
  | F (f, ts) -> 
      let u = F (f, [ tcap rs t | t <- ts ]) in 
      if List.exists (fun (l, _) -> unifiable u l) rs then
        V (fresh ())
      else
        u

let intersect xs ys = List.exists (fun x -> List.mem x ys) xs

let joinable1 n rs s t = 
  let ss = reachable n rs [s]
  and ts = reachable n rs [t] in
  intersect ss ts

let joinable n rs es =
  List.for_all (fun (s, t) -> joinable1 n rs s t) es

(** Confluence criteria *)

(* Knuth and Bendix' criterion (1970) *)
let kb70 rs = 
  let cp = cp rs in
  if not (Rsn.not_sn rs []) && 
     Sn.sn !tt rs [] = YES &&
     List.for_all (fun (s, t) -> nf rs s = nf rs t) cp
  then
    KH12 (rs, [], cp, HM11 ([], [], []))
  else
    raise Failed

(* Hirokawa and Middeldorp's criterion (JAR 2011) *)
let hm11 rs = 
  let p (s, t) = s <> t in
  let cps = cps ~p rs in
  let cp = cp rs in
  if Rules.left_linear rs && 
     not (Rsn.not_sn cps rs) && 
     joinable !wcrk rs cp &&
     Sn.sn !tt cps rs = YES
  then
    HM11 (rs, cps, cp)
  else 
    raise Failed

(* Van Oostrom's rule labeling (RTA 2008) *)

let rlv rs = 
  match Rlv.solve ~tool:!smt !rlvk rs with
  | Some a -> VO08 (rs, a)
  | None -> raise Failed

let rlc rs =
  match Rl.solve  ~tool:!smt !rlck rs with
  | Some a -> VO08 (rs, a)
  | None -> raise Failed

(* Non-confluence check *)

let unjoinable_by_tcap rs s t = 
  not (Subst.unifiable (tcap rs s) (tcap rs t))

(*
let unjoinable_by_closure k rs s t = 
  let ss1 = reachable k rs [s]
  and ts1 = reachable k rs [t] in
  let ss2 = reachable 1 rs ss1
  and ts2 = reachable 1 rs ts1 in
  Listset.subset ss2 ss1 &&
  Listset.subset ts2 ts1 && 
  not (intersect ss2 ts2)

let unjoinable k rs s t =
  unjoinable_by_tcap rs s t ||
  unjoinable_by_closure 1 rs s t
*)

let rec ground = function
  | V x -> F (x, [])
  | F (f, ts) -> F (f, [ ground t | t <- ts ])

let critical_peak rs =
  [ ground (substitute mu l2) | (_, _), _, (l2, _), mu <- overlap rs ]

let non_confluent k rs =
  let f t = pairwise_find (unjoinable_by_tcap rs) (reachable k rs [t]) in
  try_find f (critical_peak (rs @ [ r, l | l, r <- rs]))

(* Klein and Hirokawa's criterion (LPAR 2012) *)

(* joinability of S-overlap *)

let kh12_overlap_aux rule1 rule2 =
  let l1, r1 = Rule.rename rule1
  and l2, r2 = Rule.rename rule2 in
  [ (l1, r1), p, (l2, r2), mgu (subterm_at p l2) l1
  | p <- function_positions l2; 
    unifiable (ren l1) (ren (subterm_at p l2));
    p <> [] || not (Rule.variant (l1, r1) (l2, r2)) ] 

let kh12_overlap rs =
  Listset.unique [ o | a <- rs; b <- rs; o <- kh12_overlap_aux a b ]

let kh12_cp rs = [ cp_of_overlap o | o <- kh12_overlap rs ]

let kh12 cr rs ss =
  let ecp =
    try kh12_cp rs with Not_unifiable -> raise Failed in
  if sno rs ss && 
     joinable !wcrk (rs @ ss) ecp &&
     Sn.sn !tt rs ss = YES
  then
    KH12 (rs, ss, ecp, cr ss)
  else
    raise Failed

let kh12_auto cr rs =
  let f ss = kh12 cr (Listset.diff rs ss) ss in
  try_find f [ ss | ss <- Listx.power rs; ss <> [] && not (Listset.subset rs ss) ]

(* Jouannaud and Kirchner's criterion (1986) *)

let jk86 rr =
  match Ac_rewriting.find_ac rr with
  | None -> raise Failed
  | Some (rs, ac_sym, ac, ac_all) ->
      if not (Sn.sn !actt ~ac_sym rs [] = YES) then 
	raise Failed
      else 
	let cps = Rules.flatten ac_sym (
            Ac_overlap.cp  ac_sym rs
          @ Ac_overlap.cp2 ac_sym rs ac_all ) in
	let cps_nf =
           [ Ac_rewriting.nf ac_sym rs s,
             Ac_rewriting.nf ac_sym rs t 
           | s, t <- cps ] in
	let nj =
          [ s,t | s,t <- Variant.rename_rules cps_nf; 
                  not (Ac_rewriting.equal ac_sym s t) ] in
        match nj with
        | [] -> JK86 (rr, Variant.rename_rules cps)
        | (s, t) :: _ -> raise (Disproved (s, t))

let jk86_maybe rr = 
  try jk86 rr with Disproved (_, _) -> raise Failed

let rec confluent1 rs = 
  try_find (fun f -> f rs) [ kb70; hm11; rlv; rlc; jk86_maybe; kh12_auto confluent1]

let confluent rs = 
  try_find (fun f -> f rs) [ kb70; hm11; rlv; rlc; jk86; kh12_auto confluent1]

let options = align [ 
  "-t",  Float  (fun x -> t := Some x), "<f> second timeout"; 
  "-tt", String (fun x -> tt := x),
     sprintf "<tool> termination tool (default: %s)" !tt;
  "-actt", String (fun x -> actt := x),
     sprintf "<tool> AC termination tool (default: %s)" !actt;
  "-smt", String (fun x -> smt := x),
     sprintf "<tool> QF_NIA SMT solver (default: %s)" !smt;
]

(* main *)
     
let short_usage = "Usage: saigawa [options] <file>\n"
let usage = "Confluence tool Saigawa v1.4\n" ^ short_usage ^ "Options are:"

let () = 
  parse options (fun it -> filenames := !filenames @ [it] ) usage;
  if List.length !filenames <> 1 then
    (eprintf "%s%!" short_usage; exit 1)
  else
    let filename = List.hd !filenames in
    let f_timeout () = printf "TIMEOUT@." in
    let f_prove () =
      let rs = Read.read_trs filename in
      if not (Rules.variable_condition rs) then
	(eprintf "Cannot handle TRS with extra variables.@."; exit 1);
      try 
        let s, t = non_confluent !wcrk rs in
        printf "NO@.@.%a@." print_disproof (rs, s, t)
      with Failed ->
	try 
          let proof = confluent rs in
          printf "YES@.@.%a@." print_proof proof
        with
        | Disproved (s, t) -> printf "NO@.@.%a@." print_disproof (rs, s, t)
        | Failed -> printf "MAYBE@."
    in
    Alarm.try_with_timeout ?timeout:!t f_timeout f_prove ()
